/*
 * Race track generator
 *
 * leto/calodox, 02.2002
 */

#include <string.h>	// for memcopy
#include <stdlib.h> // for rand()
#include <stdio.h> // for printf (debugging)
#include <assert.h>

#include "ezspline.h"
#include "trackgen.h"
#include <math.h> // for cos/sin

//#define HARD_CODED_KEYS

#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795
#endif

// x,y,z,roll, width
typedef float SplineKey[5];

#define MAX_VERTICES (20*60)
static SplineKey trackVertices[MAX_VERTICES];


#ifdef HARD_CODED_KEYS

static SplineKey keys[] = {
	{  0, 0,  0, 0, 2.5f },
	{ 20, 1,  6, 0, 2.5f },
	{ 20, 2, 20, 0, 1.3f },
	{ 15, 5, 25, 0, 1.3f },
	{ 15, 2, 10, 0, 1.1f },
	{  2, 0, 10, 0, 1.0f }
};
int nbKeys = sizeof(keys) / sizeof(SplineKey);

#else

#define MAX_KEYS 128
static SplineKey keys[MAX_KEYS];
int nbKeys = 0;

#endif

#define NB_SECTION_PTS 4
static float section[NB_SECTION_PTS][2] = {
	{ -1.0f,  0.3f },
	{ -0.9f,     0 },
	{  0.9f,     0 },
	{  1.0f,  1.3f }
};


static void addVect(float *a, float *b, float *result) {
	for (int i=0; i<3; ++i)
		result[i]=a[i]+b[i];
}

static void subVect(float *a, float *b, float *result) {
	for (int i=0; i<3; ++i)
		result[i]=a[i]-b[i];
}

static float dotProd(float *v1, float *v2) {
	return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
}

static void crossProd( float *v1, float *v2, float *res ) {
   res[0] = v1[1] * v2[2] - v1[2] * v2[1];
   res[1] = v1[2] * v2[0] - v1[0] * v2[2];
   res[2] = v1[0] * v2[1] - v1[1] * v2[0];
}

static float length(float *v) {
	return sqrt( v[0]*v[0] + v[1]*v[1] + v[2]*v[2] );
}

static void normalize(float *v) {
	float l = length(v);
	if (l !=0 ) {
		l = 1/l;
		v[0] *= l;
		v[1] *= l;
		v[2] *= l;
	}
}

void TrackGenerator::generateTrack() {
	int v=0;

	int nbVert = nbKeys * nbSeg;

	assert(nbVert < MAX_VERTICES);

	// first pass: generate all
	// track vertices
	for (int i=0; i<nbKeys; ++i) {
		float *p = &keys[(nbKeys+i-1)%nbKeys][0];
		float *a = &keys[i][0];
		float *b = &keys[(i+1)%nbKeys][0];
		float *n = &keys[(i+2)%nbKeys][0];

		for (int j=0; j<nbSeg; j++) {

			splineCompute( (float) j / (float)nbSeg, 5, p, a, b, n, 
				&trackVertices[v][0]);
			v++;
		}
	}

	assert(v==nbVert);

	// second pass: generate segments
	float *a = &trackVertices[nbVert-1][0];
	float *prev = &trackVertices[nbVert-2][0];
	float lastPts[NB_SECTION_PTS][3];

	for (int j=0; j<=nbVert; ++j) {
		float *next = &trackVertices[j%nbVert][0];

		float s1[2], s2[2];
		
		s1[0] = a[0] - prev[0];
		s1[1] = a[2] - prev[2];
		s2[0] = next[0] - a[0];
		s2[1] = next[2] - a[2];

		float s1len = sqrt(s1[0]*s1[0] + s1[1]*s1[1]);
		float s2len = sqrt(s2[0]*s2[0] + s2[1]*s2[1]);

		for (int k=0; k<2; ++k) {
			s1[k]/=s1len;
			s2[k]/=s2len;
		}

		// Compute the angle at 'a' position 
		float ca = s1[0]*s2[0] + s1[1]*s2[1];
		float angle = (M_PI-acos(ca))/2;

		// check for the sign
		float d[3], dprev[3], tmp[3], inside[3];
		subVect(next, a, d);
		subVect(a, prev, dprev);
		crossProd(d, dprev, tmp);
		if (length(tmp) < 0.1) {
			inside[0] = d[2];
			inside[2] = -d[0];
		} else {
			crossProd(tmp, dprev, inside);
		}
		inside[1]=0;
		normalize(inside);


		// rotate s2 with this angle
		float preRoll[2];
		preRoll[0] = (s2[0]*cos(angle) - s2[1]*sin(angle));
		preRoll[1] = (s2[0]*sin(angle) + s2[1]*cos(angle));
	
		float r[3];
		r[0] = preRoll[0];
		r[1] = 0;
		r[2] = preRoll[1];

		bool rightTurn = false;
		if (dotProd(r, inside) < 0) {

			// make 'inside' point to the
			// right side of the track
			for (int k=0; k<3; k++)
				inside[k] = - inside[k];
		}

		if (dotProd(d,inside) >0) {
			rightTurn = true;
		}

		float n[3];
		crossProd(r, d, n);
		normalize(n);

		int p;
		float pts[NB_SECTION_PTS][3];

		for (p=0; p < NB_SECTION_PTS; ++p) {
			for (int l=0; l<3;l++)
				pts[p][l] = a[l] + section[p][0] * inside[l] *a[4] + section[p][1]*n[l];
		}

		if (j) {

			// triangle generation
			for (p=0; p < NB_SECTION_PTS -1; ++p) {
				int type = 0;

				// compute the type of the triangle
				switch (p) {
				case 0:
					if (rightTurn && ca < curveTrig)
						type = 2;
					else 
						type = 1;
					break;
				case NB_SECTION_PTS-2:
					if (!rightTurn && ca < curveTrig)
						type = 2;
					else 
						type = 1;
					break;
				default:
					type = 0;
					if (j%nbSeg == 0) type = 5;
				}


				addTriangle(pts[p], lastPts[p], pts[p+1], type);
				addTriangle(pts[p+1], lastPts[p], lastPts[p+1], type);
			}
		}

		memcpy(lastPts, pts, sizeof(lastPts));
		prev = a;
		a = next;
	}
}

#ifdef HARD_CODED_KEYS
void TrackGenerator::generateKeys(int level) {
}

#else
void TrackGenerator::ruleA(int level) {
	if (level < 0) level =0;

	if (level == 0) {
		advance();
		putKey();
	} else {
		float r = ((float)rand())/((float)RAND_MAX);

		if (r<.3) {
			ruleA(level-2);
			turn(-M_PI/3);
			ruleA(level-1);
			turn(2*M_PI/3);
			ruleA(level-1);
			turn(-M_PI/3);
			ruleA(level-2);
		} /*else if (r < .6) {
			ruleA(level-2);
			turn(M_PI/3);
			ruleA(level-1);
			turn(-2*M_PI/3);
			ruleA(level-1);
			turn(M_PI/3);			
			ruleA(level-2);
		} */
		else if (r < .6) {
			turn(-M_PI/3);
			ruleA(level-1);
			turn(2*M_PI/3);
			ruleA(level-1);
			turn(-M_PI/3);
			ruleA(level-2);
			ruleA(level-2);
		}
		else {
			ruleA(level-2);
			ruleA(level-1);
			ruleA(level-2);
		}
	}
}

void TrackGenerator::generateKeys(int level) {
	nbKeys = 0;
	memset(position, 0, sizeof(position));

#ifndef NDEBUG
	printf("Starting generation of level %d\n", level);
#endif
	angle=0;

	ruleA(0);
	ruleA(level);
	turn(M_PI/2);
	ruleA(level);
	turn(M_PI/2);
	ruleA(level);
	ruleA(0);
	turn(M_PI/2);
	ruleA(level);
	turn(M_PI/2);

#ifndef NDEBUG
	printf("%d keys generated.\n", nbKeys);
#endif
}

void TrackGenerator::advance() {

	position[0] += sin(angle) * keyDist;
	position[2] += cos(angle) * keyDist;
}

void TrackGenerator::putKey() {
	assert(nbKeys < MAX_KEYS -1);

	keys[nbKeys][0] = position[0];
	keys[nbKeys][1] = 0;
	keys[nbKeys][2] = position[2];
	keys[nbKeys][3] = 0;
	keys[nbKeys][4] = 1;
	printf("key at %5.2f, %5.2f\n", position[0], position[2]);

	nbKeys++;
}

#endif
